/*
 * OOP class of RGTP connections
 * header file for per-connection object datatype
 * and external RGTP C functions
 */

#ifndef RGTPRG_H
#define RGTPRG_H

#define SECRET_MAXBYTES 128

typedef struct {
  int linenumber;
  unsigned long cookie;
  time_t timestamp;
} RG_MarkInItem;

/* These must all be in the same order. */
#define RG_IET_RGTPCHARS       "ICRMEF"
#define RG_IET_DEFINLINESCHARS "*+ MEC"
#define RG_IET_WORDS  "newitem", "continuation", "reply", "motd", "edit", "continued"
typedef enum {
  RG_IET_NewItem, RG_IET_Continuation, RG_IET_Reply,
  RG_IET_MoTD, RG_IET_Edit, RG_IET_Continued,
  RG_IET_Max
} RG_IndexEntryType;

typedef struct {
  unsigned long cookie;
  time_t timestamp;
  RG_IndexEntryType type;
  char itemid[ITEMID_LEN+1];
  char userid[USERID_MAXLEN+1];
  char subject[SUBJECTININDEX_MAXLEN+1];
  int lastentrythisitem; /* bitmask, indexed by RG_IndexEntryType */
} RG_IndexEntry;

typedef struct {
  void *data;
  int allocd, used;
} RG_EArray;

typedef struct {
  enum { RG_Status_OK, RG_Status_NeedReconnect, RG_Status_Broken } status;
  enum RG_AccessLevels {
    RG_Access_None, RG_Access_Read, RG_Access_Write, RG_Access_Edit
  } access, /* not valid if status is broken */
    wantaccess; /* -1 means default */
  FILE *file; /* not open if status is broken or needreconnect */
  char *host;
  int port; /* network byte-order */
  char *user; /* null means want to register */
  int md5secretbytes;
  unsigned char md5secret[SECRET_MAXBYTES];
  int id, debug, regudone, logindone;
  Tcl_DString itemtext;
  RG_EArray marks;
  int itemvalid;
  int editing, editlocked;
  RG_EArray indexdata;
  RG_EArray indexorder;
  int indexvalid; /* 1=yes, 2=all */
  unsigned long indexsince;
  char contitem[ITEMID_LEN+1];
} RG_Connection;

typedef struct {
  const char *key;
  const char *value;
} RG_KeyValuePair;

/* from RGlogin.c */
int RG_Login(Tcl_Interp*, RG_Connection*);
int RG_Reconnect(Tcl_Interp*, RG_Connection*);
int RG_HandleAuth(Tcl_Interp *interp, RG_Connection *connP,
                  Tcl_DString *bannerP, int wantaccess, int closeifbad);

/* from RGmanage.c */
void RG_EArrayInit(RG_EArray*);
void RG_EArrayFree(RG_EArray*);
void RG_EArrayReset(RG_EArray*);
void *RG_EArrayAppend(RG_EArray*, size_t);
void RG_EArrayEnsureSize(RG_EArray*, int, size_t);
void RG_Break(RG_Connection*);
int RG_Flush(Tcl_Interp*,RG_Connection*);
int RG_StartIntIO(Tcl_Interp*,int fd);
int RG_EndIntIO(Tcl_Interp*,RG_Connection*);
void RG_Violation(RG_Connection*);
void RG_DeleteCallback(ClientData);
int RG_CommandResponse(Tcl_Interp*, RG_Connection*, Tcl_DString*, char*, ...);
int RG_CommandResponseRetry(Tcl_Interp*, RG_Connection*, Tcl_DString*, char*, ...);
int RG_ReadLine(Tcl_Interp*, RG_Connection*, Tcl_DString*);
int RG_ReadLineEcho(Tcl_Interp*, RG_Connection*, Tcl_DString*);
int RG_TransmitData(Tcl_Interp*, RG_Connection*, char *grogname,
                    char *text, const RG_KeyValuePair *errortable);
int RG_ReadData(Tcl_Interp*, RG_Connection*, Tcl_DString*);
char *RG_CommandName(RG_Connection*);
int RG_AccessOK(Tcl_Interp*, RG_Connection*, int);
int RG_GetAccessLevel(Tcl_Interp*, char *value, int *rhere);
int RG_GetMD5Secret(Tcl_Interp*, RG_Connection*, char *s);
void RG_DeleteCommand(Tcl_Interp*, RG_Connection*);

/* from RGcontrib.c */
int RG_SubmitError(Tcl_Interp*, Tcl_DString*, const RG_KeyValuePair*);
int RG_CheckSubject(Tcl_Interp*, char *value, char **startp, int *endp,
                    char *whatcode, char *whatstring, int maxlen);

/* from RGopen.c */
int RGTP_TclInit(Tcl_Interp*);
int RG_ConnectionCmd(ClientData, Tcl_Interp*, int, char **argv);
int RG_OpenCmd(ClientData clientData, Tcl_Interp*, int, char **argv);
int RG_CompareSeqsCmd(ClientData, Tcl_Interp*, int, char **argv);

/* from RGread.c */
int RG_Compare_WrappingLongs(unsigned long val, unsigned long ref);

/* subcommands from RGcontrib.c */
int RG_Object_ReplyCmd(Tcl_Interp*, RG_Connection*, int, char**);
int RG_Object_ContributeCmd(Tcl_Interp*, RG_Connection*, int, char**);
int RG_Object_ContinueCmd(Tcl_Interp*, RG_Connection*, int, char**);

/* subcommands from RGedit.c */
int RG_Object_EditLockCmd(Tcl_Interp*, RG_Connection*, int, char**);
int RG_Object_EditUnlockCmd(Tcl_Interp*, RG_Connection*, int, char**);
int RG_Object_EditCmd(Tcl_Interp*, RG_Connection*, int, char**);
int RG_Object_EditDoneCmd(Tcl_Interp*, RG_Connection*, int, char**);
int RG_Object_EditIndexCmd(Tcl_Interp*, RG_Connection*, int, char**);
int RG_Object_WithdrawCmd(Tcl_Interp*, RG_Connection*, int, char**);
int RG_Object_SetMoTDCmd(Tcl_Interp*, RG_Connection*, int, char**);
int RG_Object_UDBMCmd(Tcl_Interp*, RG_Connection*, int, char**);
int RG_Object_KillServerCmd(Tcl_Interp*, RG_Connection*, int, char**);
int RG_Object_RestartServerCmd(Tcl_Interp*, RG_Connection*, int, char**);

/* subcommands from RGindex.c */
int RG_Object_IndexInfoCmd(Tcl_Interp*, RG_Connection*, int, char**);
int RG_Object_IndexLinesCmd(Tcl_Interp*, RG_Connection*, int, char**);
int RG_Object_IndexEntryCmd(Tcl_Interp*, RG_Connection*, int, char**);
int RG_Object_IndexModifyCmd(Tcl_Interp*, RG_Connection*, int, char**);
int RG_Object_IndexCmd(Tcl_Interp*, RG_Connection*, int, char**);

/* subcommands from RGread.c */
int RG_Object_EditLogCmd(Tcl_Interp*, RG_Connection*, int, char**);
int RG_Object_MoTDCmd(Tcl_Interp*, RG_Connection*, int, char**);
int RG_Object_ItemDataCmd(Tcl_Interp*, RG_Connection*, int, char**);
int RG_Object_ItemMarkCmd(Tcl_Interp*, RG_Connection*, int, char**);
int RG_Object_ItemCmd(Tcl_Interp*, RG_Connection*, int, char**);
int RG_Object_ItemStatCmd(Tcl_Interp*, RG_Connection*, int, char**);

/* subcommands from RGopen.c */
int RG_Object_DebugCmd(Tcl_Interp*, RG_Connection*, int, char**);
int RG_Object_StatusCmd(Tcl_Interp*, RG_Connection*, int, char**);
int RG_Object_NoOpCmd(Tcl_Interp*, RG_Connection*, int, char**);
int RG_Object_CloseCmd(Tcl_Interp*, RG_Connection*, int, char**);
int RG_Object_DestroyCmd(Tcl_Interp*, RG_Connection*, int, char**);
int RG_Object_AccessLevelCmd(Tcl_Interp*, RG_Connection*, int, char**);
int RG_Object_RegisterCmd(Tcl_Interp*, RG_Connection*, int, char**);

int RG_Object_RawCommandResponseCmd(Tcl_Interp *interp, RG_Connection *connP,
                                    int argc, char **argv);
int RG_Object_RawReadLineCmd(Tcl_Interp *interp, RG_Connection *connP,
                             int argc, char **argv);
int RG_Object_RawReadDataCmd(Tcl_Interp *interp, RG_Connection *connP,
                             int argc, char **argv);
int RG_Object_RawTransmitDataCmd(Tcl_Interp *interp, RG_Connection *connP,
                                 int argc, char **argv);

#endif
