/*
 * OOP class of RGTP connections
 * implements editorial commands
 *   $conn editlock
 *      return is list, one of   OK
 *                               LOCKED userid
 *   $conn udbm
 *      return is text returned by udbm command
 *   $conn editunlock
 *   $conn edit itemid
 *      return is list, one of   OK  (item text is in itemdata)
 *                               NOITEM reason
 *   $conn editindex
 *      return is                OK  (index text is in itemdata)
 *   $conn editdone [reason newtext]
 *      without reason & newtext indicates abort edit
 *      return is empty for abort, otherise list, one of
 *                               OK
 *                               FORMAT TEXT|REASON reason
 *                               TOOLATE
 *                               NOITEM reason
 *   $conn withdraw itemid reason
 *      return is list, one of   OK sequence
 *                               NOITEM reason
 *   $conn setmotd newmotd
 *      return is list, one of   OK sequence
 *                               FORMAT TEXT reason
 */

#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <assert.h>
#include <limits.h>
#include <time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <ctype.h>

#include "tcl.h"
#include "tclExtend.h"
#include "rgtp.h"

#include "fixup.h"
#include "md5.h"
#include "RG.h"

static const RG_KeyValuePair editerrors[] = {
  { "423 ", "FORMAT TEXT"   },
  { "424 ", "FORMAT REASON" },
  { "410 ", "NOITEM"        },
  {  0                      }
};

static const RG_KeyValuePair motserrors[] = {
  { "423 ", "FORMAT TEXT"   },
  {  0                      }
};

static int checkbeginedit(Tcl_Interp *interp, RG_Connection *connP) {
  if (RG_AccessOK(interp,connP,RG_Access_Edit) != TCL_OK) return TCL_ERROR;
  if (connP->editlocked) return TCL_OK;
  interp->result= "must use editlock before doing an edit operation";
  return TCL_ERROR;
}

int RG_Object_EditLockCmd(Tcl_Interp *interp, RG_Connection *connP,
                          int argc, char **argv) {
  Tcl_DString ds;
  char *p;
  int r;

  if (argc != 1) { interp->result= "wrong # args"; return TCL_ERROR; }
  if (RG_AccessOK(interp,connP,RG_Access_Edit) != TCL_OK) return TCL_ERROR;
  if (connP->editlocked) { interp->result= "already locked"; return TCL_ERROR; }
  r= RG_CommandResponseRetry(interp,connP,&ds, "EDLK");
  if (r != TCL_OK) return r;
  p= Tcl_DStringValue(&ds);
  if (!strncmp(p,"411 ",4)) {
    strcpy(interp->result,"LOCKED");
    p= strchr(p+4,' '); if (p) *p=0;
    Tcl_AppendElement(interp, Tcl_DStringValue(&ds)+4);
    Tcl_DStringFree(&ds);
    return TCL_OK;
  }
  if (!strncmp(Tcl_DStringValue(&ds),"200 ",4)) {
    Tcl_DStringFree(&ds);
    connP->editlocked= 1;
    interp->result= "OK";
    return TCL_OK;
  }
  Tcl_AppendResult(interp,
                   "EDLK command rejected by server:\n",
                   Tcl_DStringValue(&ds), (char*)0);
  Tcl_DStringFree(&ds);
  return TCL_ERROR;
}

int RG_Object_EditUnlockCmd(Tcl_Interp *interp, RG_Connection *connP,
                            int argc, char **argv) {
  Tcl_DString ds;
  int r;
  
  if (argc != 1) { interp->result= "wrong # args"; return TCL_ERROR; }
  if (RG_AccessOK(interp,connP,RG_Access_Edit) != TCL_OK) return TCL_ERROR;
  if (connP->editing) {
    interp->result= "cannot unlock before edit finished"; return TCL_ERROR;
  }
  if (!connP->editlocked) {
    interp->result= "not currently locked, cannot unlock"; return TCL_ERROR;
  }
  r= RG_CommandResponse(interp,connP,&ds, "EDUL");
  if (r != TCL_OK) return r;
  if (!strncmp(Tcl_DStringValue(&ds),"200 ",4)) {
    connP->editlocked= 0; Tcl_DStringFree(&ds); return TCL_OK;
  } else if (!strncmp(Tcl_DStringValue(&ds),"481 ",4)) {
    RG_Break(connP); connP->status= RG_Status_NeedReconnect; return TCL_OK;
  }
  Tcl_AppendResult(interp,
                   "EDUL command rejected by server:\n",
                   Tcl_DStringValue(&ds), (char*)0);
  Tcl_DStringFree(&ds);
  return TCL_ERROR;
}

int RG_Object_EditCmd(Tcl_Interp *interp, RG_Connection *connP,
                      int argc, char **argv) {
  /*
   * $conn edit itemid
   *    return is list, one of   OK  (item text is in itemdata)
   *                             NOITEM reason
   *                             LOCKED userid reason
   */
  char buf[20], *p;
  Tcl_DString ds;
  int r;

  if (argc != 2) { interp->result= "wrong # args"; return TCL_ERROR; }
  if (checkbeginedit(interp,connP) != TCL_OK) return TCL_ERROR;
  if (strlen(argv[1]) != ITEMID_LEN) {
    interp->result= "item-id is wrong length"; return TCL_ERROR;
  }
  strcpy(buf,"EDIT "); strcat(buf,argv[1]);

  connP->contitem[0]= 0;
  r= RG_CommandResponseRetry(interp,connP,&ds,buf);
  if (r != TCL_OK) return r;
  p= Tcl_DStringValue(&ds);
  if (!strncmp(p,"410 ",4)) {
    strcpy(interp->result,"NOITEM");
    Tcl_AppendElement(interp, p+4);
    Tcl_DStringFree(&ds);
    return TCL_OK;
  }
  if (strncmp(Tcl_DStringValue(&ds),"250 ",4)) {
    Tcl_AppendResult(interp,
                     "server didn't provide item to be edited ", argv[1],
                     " on request:\n", p, (char*)0);
    Tcl_DStringFree(&ds);
    return TCL_ERROR;
  }
  Tcl_DStringFree(&ds);
  r= RG_ReadData(interp,connP,&ds);
  if (r != TCL_OK) return r;
  Tcl_DStringFree(&connP->itemtext);
  RG_EArrayReset(&connP->marks);
  connP->itemtext= ds;
  connP->itemvalid= 1;
  strcpy(interp->result,"OK");
  connP->editing= 1;
  return TCL_OK;
}

int RG_Object_EditIndexCmd(Tcl_Interp *interp, RG_Connection *connP,
                           int argc, char **argv) {
  /*
   * $conn editindex
   *    return is list, one of   OK  (index text is in itemdata)
   *                             LOCKED userid
   */
  char *p;
  Tcl_DString ds;
  int r;

  if (argc != 1) { interp->result= "wrong # args"; return TCL_ERROR; }
  if (checkbeginedit(interp,connP) != TCL_OK) return TCL_ERROR;
  connP->contitem[0]= 0;
  r= RG_CommandResponseRetry(interp,connP,&ds,"EDIX");
  if (r != TCL_OK) return r;
  p= Tcl_DStringValue(&ds);
  if (!strncmp(p,"411 ",4)) {
    strcpy(interp->result,"LOCKED ");
    p= strchr(p+4,' '); if (p) *p=0;
    Tcl_AppendElement(interp, p+4);
    Tcl_DStringFree(&ds);
    return TCL_OK;
  }
  if (strncmp(Tcl_DStringValue(&ds),"250 ",4)) {
    Tcl_AppendResult(interp,
                     "server didn't provide index to be edited ", argv[1],
                     " on request:\n", p, (char*)0);
    Tcl_DStringFree(&ds);
    return TCL_ERROR;
  }
  Tcl_DStringFree(&ds);
  r= RG_ReadData(interp,connP,&ds);
  if (r != TCL_OK) return r;
  Tcl_DStringFree(&connP->itemtext);
  RG_EArrayReset(&connP->marks);
  connP->itemtext= ds;
  connP->itemvalid= 1;
  strcpy(interp->result,"OK");
  connP->editing= 1;
  return TCL_OK;
}

int RG_Object_EditDoneCmd(Tcl_Interp *interp, RG_Connection *connP,
                          int argc, char **argv) {
  /*
   * $conn editdone [reason newtext]
   *    without reason & newtext indicates abort edit
   *    return is empty for abort, otherise list, one of
   *                             OK
   *                             FORMAT TEXT|REASON reason
   *                             TOOLATE
   *                             NOITEM reason
   */
  char buf[TXRXLINE_MAXLEN+20], *p, *ep;
  Tcl_DString ds;
  int r, l;
  unsigned long gsn;

  if (!connP->editing) { interp->result= "not editing"; return TCL_ERROR; }
  if (argc == 1) {
    r= RG_CommandResponse(interp,connP,&ds,"EDAB");
    if (r != TCL_OK) return r;
    p= Tcl_DStringValue(&ds);
    if (!strncmp(p,"481 ",4)) {
      RG_Break(connP);
      connP->status= RG_Status_NeedReconnect;
    } else if (strncmp(p,"200 ",4)) {
      Tcl_AppendResult(interp,
                       "server failed/refused to abort edit correctly:\n",
                       p, (char*)0);
      Tcl_DStringFree(&ds);
      RG_Violation(connP);
      return TCL_ERROR;
    }
    Tcl_DStringFree(&ds);
    connP->editing= 0;
    return TCL_OK;
  } else if (argc == 3) {
    r= RG_CheckSubject(interp,argv[1],&p,&l,"REASON","reason",TXRXLINE_MAXLEN-5);
    if (r == TCL_RETURN) return TCL_OK; else if (r != TCL_OK) return r;
    
    r= RG_TransmitData(interp,connP,0,argv[2],editerrors);
    if (r == TCL_RETURN) return TCL_OK;
    if (r != TCL_OK) return r;
    if (!connP->editing) { interp->result= "TOOLATE"; return TCL_OK; }
    strcpy(buf,"EDCF "); memcpy(buf+5,p,l); buf[5+l]= 0;
    r= RG_CommandResponse(interp,connP,&ds,buf);
    if (r != TCL_OK) return r;
    if (RG_SubmitError(interp,&ds,editerrors)) return TCL_OK;
    connP->editing= 0;
    p= Tcl_DStringValue(&ds);
    if (strncmp(p,"220 ",4)) {
      Tcl_AppendResult(interp,
                        "server failed/refused to confirm edit correctly:\n",
                       p, (char*)0);
      Tcl_DStringFree(&ds);
      RG_Violation(connP);
      return TCL_ERROR;
    }
    gsn= strtoul(p+4,&ep,16);
    if (ep != p+12 || (*ep && *ep!=' ')) {
      Tcl_AppendResult(interp,
                       "server sent bad 220 reply to EDCF update:\n", p,
                       (char*)0);
      Tcl_DStringFree(&ds);
      return TCL_ERROR;
    }
    Tcl_DStringFree(&ds);
    sprintf(interp->result,"OK %08lx",gsn);
    return TCL_OK;
  } else {
    interp->result= "wrong # args"; return TCL_ERROR;
  }
}

int RG_Object_WithdrawCmd(Tcl_Interp *interp, RG_Connection *connP,
                          int argc, char **argv) {
  /*
   * $conn withdraw itemid reason
   *    return is list, one of   OK sequence
   *                             NOITEM reason
   *                             LOCKED userid
   */
  char buf[TXRXLINE_MAXLEN+20], *p, *ep, *reas;
  Tcl_DString ds;
  int r, l;
  unsigned long gsn;

  if (argc != 3) { interp->result= "wrong # args"; return TCL_ERROR; }
  if (checkbeginedit(interp,connP) != TCL_OK) return TCL_ERROR;
  if (strlen(argv[1]) != ITEMID_LEN) {
    interp->result= "item-id is wrong length"; return TCL_ERROR;
  }
  r= RG_CheckSubject(interp,argv[2],&reas,&l, "REASON","reason",TXRXLINE_MAXLEN-5);
  if (r == TCL_RETURN) return TCL_OK; else if (r != TCL_OK) return r;
  strcpy(buf,"EDIT "); strcat(buf,argv[1]);

  r= RG_CommandResponseRetry(interp,connP,&ds,buf);
  if (r != TCL_OK) return r;
  p= Tcl_DStringValue(&ds);
  if (!strncmp(p,"410 ",4)) {
    strcpy(interp->result,"NOITEM");
    Tcl_AppendElement(interp, p+4);
    Tcl_DStringFree(&ds);
    return TCL_OK;
  }
  if (strncmp(Tcl_DStringValue(&ds),"250 ",4)) {
    Tcl_AppendResult(interp,
                     "server didn't provide item to be withdrawn ", argv[1],
                     " on request:\n", p, (char*)0);
    Tcl_DStringFree(&ds);
    return TCL_ERROR;
  }
  Tcl_DStringFree(&ds);
  
  r= RG_ReadData(interp,connP,&ds);
  if (r != TCL_OK) return r;
  Tcl_DStringFree(&ds);
  
  strcpy(buf,"EDCF "); memcpy(buf+5,reas,l); buf[5+l]= 0;
  r= RG_CommandResponse(interp,connP,&ds,buf);
  if (r != TCL_OK) return r;
  if (RG_SubmitError(interp,&ds,editerrors)) return TCL_OK;
  p= Tcl_DStringValue(&ds);
  if (strncmp(p,"220 ",4)) {
    Tcl_AppendResult(interp,
                     "server failed/refused to confirm withdrawal correctly:\n",
                     p, (char*)0);
    Tcl_DStringFree(&ds);
    RG_Violation(connP);
    return TCL_ERROR;
  }
  gsn= strtoul(p+4,&ep,16);
  if (ep != p+12 || (*ep && *ep!=' ')) {
    Tcl_AppendResult(interp,
                     "server sent bad 220 reply to EDCF withdraw:\n", p,
                     (char*)0);
    Tcl_DStringFree(&ds);
    return TCL_ERROR;
  }
  Tcl_DStringFree(&ds);
  sprintf(interp->result,"OK %08lx",gsn);
  return TCL_OK;
}

int RG_Object_SetMoTDCmd(Tcl_Interp *interp, RG_Connection *connP,
                         int argc, char **argv) {
  /*
   * $conn setmotd newmotd
   *    return is list, one of   OK sequence
   *                             FORMAT TEXT reason
   */
  char *p, *ep;
  Tcl_DString ds;
  int r;
  unsigned long gsn;

  if (argc != 2) { interp->result= "wrong # args"; return TCL_ERROR; }
  if (RG_AccessOK(interp,connP,RG_Access_Edit) != TCL_OK) return TCL_ERROR;
  
  r= RG_TransmitData(interp,connP,"",argv[1],motserrors);
  if (r == TCL_RETURN) return TCL_OK;
  if (r != TCL_OK) return r;
  
  r= RG_CommandResponse(interp,connP,&ds,"MOTS");
  if (r != TCL_OK) return r;
  if (RG_SubmitError(interp,&ds,motserrors)) return TCL_OK;
  p= Tcl_DStringValue(&ds);
  if (strncmp(p,"220 ",4)) {
    Tcl_AppendResult(interp,
                     "server failed/refused to confirm edit correctly:\n",
                     p, (char*)0);
    Tcl_DStringFree(&ds);
    RG_Violation(connP);
    return TCL_ERROR;
  }
  gsn= strtoul(p+4,&ep,16);
  if (ep != p+12 || (*ep && *ep!=' ')) {
    Tcl_AppendResult(interp,
                     "server sent bad 220 reply to MOTS:\n", p,
                     (char*)0);
    Tcl_DStringFree(&ds);
    return TCL_ERROR;
  }
  Tcl_DStringFree(&ds);
  sprintf(interp->result,"OK %08lx",gsn);
  return TCL_OK;
}

int RG_Object_UDBMCmd(Tcl_Interp *interp, RG_Connection *connP,
                      int argc, char **argv) {
  Tcl_DString ds;
  int r;

  if (argc != 2) { interp->result= "wrong # args"; return TCL_ERROR; }
  if (RG_AccessOK(interp,connP,RG_Access_Edit) != TCL_OK) return TCL_ERROR;
  r= RG_CommandResponseRetry(interp,connP,&ds,"UDBM %s",argv[1]);
  if (r != TCL_OK) return r;
  if (strncmp(Tcl_DStringValue(&ds),"250 ",4)) {
    Tcl_AppendResult(interp, "server didn't accept our UDBM:\n",
                     Tcl_DStringValue(&ds), (char*)0);
    Tcl_DStringFree(&ds);
    return TCL_ERROR;
  }
  Tcl_DStringFree(&ds);
  r= RG_ReadData(interp,connP,&ds);
  if (r != TCL_OK) return r;
  Tcl_DStringResult(interp,&ds);
  return TCL_OK;
}

int RG_Object_KillServerCmd(Tcl_Interp *interp, RG_Connection *connP,
                            int argc, char **argv) {
  Tcl_DString ds;
  int r;
  
  if (argc != 1) { interp->result= "wrong # args"; return TCL_ERROR; }
  if (RG_AccessOK(interp,connP,RG_Access_Edit) != TCL_OK) return TCL_ERROR;
  r= RG_CommandResponseRetry(interp,connP,&ds, "KILL");
  if (r != TCL_OK) return r;
  if (!strncmp(Tcl_DStringValue(&ds),"200 ",4)) { Tcl_DStringFree(&ds); return TCL_OK; }
  Tcl_AppendResult(interp,
                   "KILL command rejected by server:\n",
                   Tcl_DStringValue(&ds), (char*)0);
  Tcl_DStringFree(&ds);
  return TCL_ERROR;
}

int RG_Object_RestartServerCmd(Tcl_Interp *interp, RG_Connection *connP,
                            int argc, char **argv) {
  Tcl_DString ds;
  int r;
  
  if (argc != 1) { interp->result= "wrong # args"; return TCL_ERROR; }
  if (RG_AccessOK(interp,connP,RG_Access_Edit) != TCL_OK) return TCL_ERROR;
  r= RG_CommandResponseRetry(interp,connP,&ds, "KILR");
  if (r != TCL_OK) return r;
  if (!strncmp(Tcl_DStringValue(&ds),"200 ",4)) { Tcl_DStringFree(&ds); return TCL_OK; }
  Tcl_AppendResult(interp,
                   "KILR command rejected by server:\n",
                   Tcl_DStringValue(&ds), (char*)0);
  Tcl_DStringFree(&ds);
  return TCL_ERROR;
}
