/*
 * OOP class of RGTP connections
 * implements index commands
 *   $conn index [seqno]
 *   $conn indexinfo [-status] [-lines] [-since]
 *   $conn indexlines [-maxuserdi n] [-motdsubject subj]
 *                    [-maxlinelen n] [-typechars key-types values]
 *   $conn indexentry [-lastentlist] n
 *   $conn indexmunge [-reset] [-reverse] [-types types] [-lastent types]
 * types is a string of characters from RG_IET_RGTPCHARS (ICRMEF)
 * or a list of words from RG_IET_WORDS (newitem, continuation, reply,
 *  motd, edit, continued),
 * or ~ or ^ and optional whitespace followed by one of those two (only
 *  where order is insignificant, ie in indexmunge -types & -lastent.
 * In indexlines -typechars values is a series of characters in the same
 * order as the key-types, giving characters to display.  If values is
 * shorter extra types not displayed.  values may not be longer.
 */

#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <assert.h>
#include <limits.h>
#include <time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <ctype.h>

#include "tcl.h"
#include "tclExtend.h"
#include "rgtp.h"

#include "fixup.h"
#include "md5.h"
#include "RG.h"

static char *typestrings[]= { RG_IET_WORDS, 0 };
static char indextypelist[]= RG_IET_RGTPCHARS;
  
static int RG_IndexInfo_StatusCmd(Tcl_Interp *interp, RG_Connection *connP) {
  Tcl_AppendElement(interp,
                    connP->indexvalid==2 ? "full" :
                    connP->indexvalid==1 ? "part" : "broken");
  return TCL_OK;
}

static int RG_IndexInfo_LinesCmd(Tcl_Interp *interp, RG_Connection *connP) {
  char buf[10];
  sprintf(buf,"%d", connP->indexvalid ? connP->indexorder.used : -1);
  Tcl_AppendElement(interp,buf);
  return TCL_OK;
}

static int RG_IndexInfo_SinceCmd(Tcl_Interp *interp, RG_Connection *connP) {
  char buf[10];
  if (connP->indexvalid==1) {
    sprintf(buf,"%08lX", connP->indexsince);
  } else {
    buf[0]= 0;
  }
  Tcl_AppendElement(interp,buf);
  return TCL_OK;
}

int RG_Object_IndexInfoCmd(Tcl_Interp *interp, RG_Connection *connP,
                           int argc, char **argv) {
  static struct {
    char *cmd;
    int (*fn)(Tcl_Interp*,RG_Connection*);
  } table[] = {
    { "-status",   RG_IndexInfo_StatusCmd  },
    { "-lines",    RG_IndexInfo_LinesCmd   },
    { "-since",    RG_IndexInfo_SinceCmd   },
    { 0 }
  }, *tp;
  int r;
  char *p;
  if (argc == 1) {
    for (tp= table; tp->cmd; tp++) {
      r= tp->fn(interp,connP);  if (r != TCL_OK) return r;
    }
  } else {
    while ((p= *++argv) != 0) {
      for (tp= table; tp->cmd && strcmp(p,tp->cmd); tp++);
      if (tp->cmd) {
        r= tp->fn(interp,connP);  if (r != TCL_OK) return r;
      } else {
        Tcl_ResetResult(interp);
        Tcl_AppendResult(interp,
                         "unknown RGTP index-info option \"",
                         p, "\": must be one of", (char*)0);
        for (tp= table; tp->cmd; tp++)
          Tcl_AppendResult(interp, " ",tp->cmd,(char*)0);
        return TCL_ERROR;
      }
    }
  }
  return TCL_OK;
}

static int get_types(Tcl_Interp *interp, char *typestring,
                     RG_IndexEntryType list[RG_IET_Max+1],
                     int *bitmaskp) {
  /* list will be terminated with RG_IET_Max, and index of that
   * sentinal will be put in *filledp (unless filledp is null).
   * bitmask is in *bitmaskp (unless null).
   * list will not be used if it is null.
   */
  int filled, listc, bitmask, invert=0, r;
  RG_IndexEntryType ta, *ietp;
  char *p, *etf, buf[2], **listv, *emsg, **ets;

  emsg= "bad type character";
  if (*typestring == '~' || *typestring == '^') {
    if (list) {
      interp->result= "inverted type string not allowed in this context";
      return TCL_ERROR;
    }
    invert= (1 << RG_IET_Max)-1;
    do { typestring++; } while (isspace(*typestring));
  }
  for (p=typestring, filled=0, bitmask=0, ietp=list;
       *p && (etf= strchr(indextypelist,*p));
       p++, filled++) {
    ta= (RG_IndexEntryType)(etf - indextypelist);
    if (bitmask & (1<<ta)) { emsg= "repeated type character"; break; }
    bitmask |= (1<<ta);
    if (ietp) *ietp++= ta;
  }
  if (*p) {
    buf[0]= *p; buf[1]= 0;
    r= Tcl_SplitList(interp,typestring,&listc,&listv);
    if (r != TCL_OK) {
      Tcl_AppendResult(interp, " / ",emsg," `",buf,"'",(char*)0); return TCL_ERROR;
    }
    for (filled=0, bitmask=0, ietp=list;
         (p= *listv) != 0;
         listv++, filled++) {
      for (ta=0, ets=typestrings;
           *ets && strcmp(p,*ets);
           ta++, ets++);
      if (!*ets) {
        Tcl_AppendResult(interp, "unknown entry type ",p, " / ",
                         emsg," `",buf,"'",(char*)0);
        ckfree(listv); return TCL_ERROR;
      }
      if (bitmask & (1<<ta)) {
        Tcl_AppendResult(interp, "repeated entry type ",p, " / ",
                         emsg," `",buf,"'",(char*)0);
        ckfree(listv); return TCL_ERROR;
      }
      bitmask |= (1<<ta);
      if (ietp) *ietp++= ta;
    }
    ckfree(listv);
  }
  bitmask ^= invert;
  if (ietp) *ietp++= RG_IET_Max;
  if (bitmaskp) *bitmaskp= bitmask;
  return TCL_OK;
}

int RG_Object_IndexLinesCmd(Tcl_Interp *interp, RG_Connection *connP,
                            int argc, char **argv) {
  RG_IndexEntry *entry, **iepp;
  char *p, *q;
  int maxuserdi=14, r, ii;
  char datebuf[50];
  char buf[ITEMID_LEN + 2 + USERID_MAXLEN + sizeof(datebuf) +
           SUBJECTININDEX_MAXLEN + 2];
  char uidbuf[USERID_MAXLEN + 2];
  char typecharlist[]= RG_IET_DEFINLINESCHARS;
  char *motdsubject= "";
  int c, maxlinelen=-1;
  RG_IndexEntryType sctypeslist[RG_IET_Max+1], *ietp;
  struct tm *tmbp;

  while ((p= *++argv) != 0 && p[0] == '-') {
    if (!strcmp(p,"-maxuserdi")) {
      p= *++argv;
      r= Tcl_GetInt(interp,p,&maxuserdi); if (r != TCL_OK) return r;
      if (maxuserdi > USERID_MAXLEN) maxuserdi= USERID_MAXLEN;
      else if (maxuserdi < 4) maxuserdi= 0;
    } else if (!strcmp(p,"-maxlinelen")) {
      p= *++argv;
      r= Tcl_GetInt(interp,p,&maxlinelen); if (r != TCL_OK) return r;
      if (maxlinelen <= 0) maxlinelen= 1;
    } else if (!strcmp(p,"-typechars")) {
      p= *++argv; q=0;
      if (p) q= *++argv;
      if (!q) {
        interp->result=
          "wrong format args: -typechars must be followed by key-types, characters.";
        return TCL_ERROR;
      }
      r= get_types(interp,p,sctypeslist,0); if (r != TCL_OK) return r;
      for (ietp= sctypeslist; *ietp != RG_IET_Max; ietp++) {
        c= *q; if (c) q++;
        typecharlist[*ietp]= c;
      }
      if (*q) {
        interp->result= "-typechars values must be same or shorter than keys";
        return TCL_ERROR;
      }
    } else if (!strcmp(p,"-motdsubject")) {
      p= *++argv;
      if (!p || strlen(p) > SUBJECTININDEX_MAXLEN) {
        sprintf(interp->result,
                 "-motdsubject must be followed by string of up to %d chars",
                SUBJECTININDEX_MAXLEN);
        return TCL_ERROR;
      }
      motdsubject= p;
    } else {
      Tcl_AppendResult(interp, "unknown option ",p," to index-lines command",
                       (char*)0);
      return TCL_ERROR;
    }
  }
  if (p) {
    interp->result= "all options to index-lines start with `-'";
    return TCL_ERROR;
  }
  if (!connP->indexvalid) {
    interp->result= "no index stored - use the index command";
    return TCL_ERROR;
  }
  for (ii=0, iepp= (RG_IndexEntry**)connP->indexorder.data;
       ii<connP->indexorder.used;
       ii++, iepp++) {
    entry= *iepp;
    if (!typecharlist[entry->type]) continue;
    if (maxuserdi) {
      if (strlen(entry->userid) > maxuserdi) {
        memcpy(uidbuf, entry->userid, maxuserdi-1);
        uidbuf[maxuserdi-1]= '>';
        uidbuf[maxuserdi]= 0;
      } else {
        strcpy(uidbuf, entry->userid);
      }
    } else {
      uidbuf[0]= 0;
    }
    tmbp= localtime(&entry->timestamp);
    if (!tmbp) {
      Tcl_ResetResult(interp);
      p= Tcl_PosixError(interp);
      Tcl_AppendResult(interp, "couldn't decompose the time: ",
                       p, (char*)0);
      return TCL_ERROR;
    }
    r= strftime(datebuf, sizeof(datebuf)-1, "%H.%M %a %d %b", tmbp);
    if (r >= sizeof(datebuf)) {
      Tcl_ResetResult(interp);
      interp->result= "date & time for index was way too long";
      return TCL_ERROR;
    }
    datebuf[sizeof(datebuf)-1]= 0;
    sprintf(buf,
            "%*s  %-*s%s %c %s",
            ITEMID_LEN, entry->itemid,
            maxuserdi+1, uidbuf,
            datebuf,
            typecharlist[entry->type],
            entry->type == RG_IET_MoTD && !entry->subject[0]
              ? motdsubject : entry->subject);
    if (strlen(buf) > maxlinelen) buf[maxlinelen]= 0;
    Tcl_AppendElement(interp,buf);
  }
  return TCL_OK;
}

int RG_Object_IndexEntryCmd(Tcl_Interp *interp, RG_Connection *connP,
                            int argc, char **argv) {
  int n, r, i, lastentlist= 0;
  Tcl_DString ds;
  RG_IndexEntry *entry;
  char buf[RG_IET_Max+10], *p;

  if (argc == 3 && !strcmp(argv[1],"-lastentlist")) {
    argc--; argv++; lastentlist= 1;
  } else if (argc != 2) {
    interp->result= "wrong # args or bad option to indexentry"; return TCL_ERROR;
  }
  if (argc != 2) { interp->result= "wrong # args"; return TCL_ERROR; }
  r= Tcl_GetInt(interp,argv[1],&n); if (r != TCL_OK) return r;
  if (!connP->indexvalid) {
    interp->result= "no index stored - use the index command";
    return TCL_ERROR;
  }
  if (n < 0 || n >= connP->indexorder.used) {
    interp->result= "index entry number out of range 0...n-1";
    return TCL_ERROR;
  }
  entry= ((RG_IndexEntry**)connP->indexorder.data)[n];
  assert((unsigned)entry->type < sizeof(typestrings)/sizeof(char*));
  Tcl_AppendElement(interp,typestrings[entry->type]);
  if (lastentlist) {
    Tcl_DStringInit(&ds);
    for (i=0; i<RG_IET_Max; i++)
      if (entry->lastentrythisitem & (1<<i))
        Tcl_DStringAppendElement(&ds,typestrings[i]);
    Tcl_AppendElement(interp,Tcl_DStringValue(&ds));
    Tcl_DStringFree(&ds);
  } else {
    for (i=0, p=buf; i<RG_IET_Max; i++)
      if (entry->lastentrythisitem & (1<<i))
        *p++= indextypelist[i];
    *p++= 0;
    Tcl_AppendElement(interp,buf);
  }
  sprintf(buf,"%08lX",entry->cookie);
  Tcl_AppendElement(interp,buf);
  sprintf(buf,"%08lX",entry->timestamp);
  Tcl_AppendElement(interp,buf);
  Tcl_AppendElement(interp,entry->itemid);
  Tcl_AppendElement(interp,entry->userid);
  Tcl_AppendElement(interp,entry->subject);
  return TCL_OK;
}

static int RG_IndexModify_ResetCmd(Tcl_Interp *interp, RG_Connection *connP,
                                   char **argv) {
  RG_IndexEntry *entry, **iepp;
  int i;
  
  for (entry= (RG_IndexEntry*)connP->indexdata.data,
       iepp= (RG_IndexEntry**)connP->indexorder.data,
       i= connP->indexdata.used;
       i >0;
       i--, iepp++, entry++) *iepp= entry;
  connP->indexorder.used= connP->indexdata.used;
  return TCL_OK;
}                           

static int RG_IndexModify_ReverseCmd(Tcl_Interp *interp, RG_Connection *connP,
                                     char **argv) {
  RG_IndexEntry **iepp, **iepq, *tempp;

  if (!connP->indexorder.used) return TCL_OK;
  for (iepp= (RG_IndexEntry**)connP->indexorder.data,
       iepq= (RG_IndexEntry**)connP->indexorder.data + connP->indexorder.used - 1;
       iepq > iepp;
       iepp++, iepq--) {
    tempp= *iepp;
    *iepp= *iepq;
    *iepq= tempp;
  }
  return TCL_OK;
}                           

static int RG_IndexModify_TypesCmd(Tcl_Interp *interp, RG_Connection *connP,
                                   char **argv) {
  RG_IndexEntry **iepp, **iepq;
  int i, mask, r, j;

  r= get_types(interp,argv[1], 0,&mask);  if (r != TCL_OK) return r;

  for (iepp= (RG_IndexEntry**)connP->indexorder.data,
       iepq= (RG_IndexEntry**)connP->indexorder.data,
       i= connP->indexorder.used, j= 0;
       i >0;
       i--, iepp++)
    if ((1 << (*iepp)->type) & mask) { *iepq++= *iepp; j++; }
  connP->indexorder.used= j;
  return TCL_OK;
}                

static int RG_IndexModify_LastEntCmd(Tcl_Interp *interp, RG_Connection *connP,
                                     char **argv) {
  RG_IndexEntry **iepp, **iepq;
  int i, mask, r, j;

  r= get_types(interp,argv[1], 0,&mask);  if (r != TCL_OK) return r;

  for (iepp= (RG_IndexEntry**)connP->indexorder.data,
       iepq= (RG_IndexEntry**)connP->indexorder.data,
       i= connP->indexorder.used, j= 0;
       i >0;
       i--, iepp++)
    if (!(mask & ~(*iepp)->lastentrythisitem)) { *iepq++= *iepp; j++; }
  connP->indexorder.used= j;
  return TCL_OK;
}                           

int RG_Object_IndexModifyCmd(Tcl_Interp *interp, RG_Connection *connP,
                             int argc, char **argv) {
  static struct {
    char *cmd;
    int (*fn)(Tcl_Interp*,RG_Connection*,char **argv);
    int argc;
  } table[] = {
    { "-reset",    RG_IndexModify_ResetCmd,    0  },
    { "-reverse",  RG_IndexModify_ReverseCmd,  0  },
    { "-types",    RG_IndexModify_TypesCmd,    1  },
    { "-lastent",  RG_IndexModify_LastEntCmd,  1  },
    { 0 }
  }, *tp;
  int r;
  char *p;
  
  if (--argc <= 0) { interp->result= "wrong # args to indexmunge"; return TCL_ERROR; }
  while (argc-- > 0) {
    p= *++argv;
    for (tp= table; tp->cmd && strcmp(p,tp->cmd); tp++);
    if (!tp->cmd) {
      Tcl_AppendResult(interp, "unknown RGTP indexmodify option `", p, "'", (char*)0);
      return TCL_ERROR;
    }
    if ((argc -= tp->argc) < 0) {
      Tcl_AppendResult(interp,
                       "insufficient args to indexmodify option ",tp->cmd,(char*)0);
      return TCL_ERROR;
    }
    r= tp->fn(interp,connP,argv);
    if (r != TCL_OK) return r;
    argv += tp->argc;
  }
  return TCL_OK;
}
  
  
int RG_Object_IndexCmd(Tcl_Interp *interp, RG_Connection *connP,
                      int argc, char **argv) {
  char buf[30];
  int level, r, i;
  unsigned long since;
  char *ep, *p, *q, *tp;
  Tcl_DString ds;
  RG_IndexEntry *entry, *centry;
  
  if (argc == 1) {
    level= 2;
    strcpy(buf,"INDX");
    since= 0;
  } else if (argc == 2) {
    since= strtoul(argv[1],&ep,16);
    if (*ep) {
      interp->result= "wrong format arg: should be 8 hex";
      return TCL_ERROR;
    }
    level= 1;
    sprintf(buf,"INDX #%08lX",since);
  } else {
    interp->result= "wrong # args";
    return TCL_ERROR;
  }
  if (RG_AccessOK(interp,connP,RG_Access_Read) != TCL_OK) return TCL_ERROR;
  connP->indexvalid= 0;
  RG_EArrayReset(&connP->indexdata);
  RG_EArrayReset(&connP->indexorder);
  r= RG_CommandResponseRetry(interp,connP,&ds,buf);
  if (r != TCL_OK) return r;
  if (strncmp(Tcl_DStringValue(&ds),"250 ",4)) {
    Tcl_AppendResult(interp, "server didn't provide index when requested:\n",
                     Tcl_DStringValue(&ds), (char*)0);
    Tcl_DStringFree(&ds);
    return TCL_ERROR;
  }
  Tcl_DStringFree(&ds);
  r= RG_ReadData(interp,connP,&ds);
  if (r != TCL_OK) return r;
  p= Tcl_DStringValue(&ds);
  while (*p) {
    q= strchr(p,'\n');
    assert(q); q++;
    if (q != p + INDEXENTRY_LEN + 1) goto x_badentry;
    entry= (RG_IndexEntry*)RG_EArrayAppend(&connP->indexdata, sizeof(RG_IndexEntry));
    entry->cookie= strtoul(p,&ep,16);
    if (ep != p+8 || *ep != ' ') goto x_badentry;
    entry->timestamp= strtoul(p+9,&ep,16);
    if (ep != p+17 || *ep != ' ') goto x_badentry;
    strncpy(entry->itemid, p+18, ITEMID_LEN); entry->itemid[ITEMID_LEN]= 0;
    i=8; tp= entry->itemid;
    if (*tp == ' ') {
      for (tp++; --i; tp++) if (*tp != ' ') goto x_badentry;
      entry->itemid[0]= 0;
    } else {
      for (; i--; tp++) if (!isalnum(*tp)) goto x_badentry;
    }
    if (p[18+ITEMID_LEN] != ' ') goto x_badentry;
    tp= p+19+ITEMID_LEN+USERID_MAXLEN;
    if (*tp != ' ') goto x_badentry;
    while (tp > p+19+ITEMID_LEN && tp[-1] == ' ') --tp;
    memcpy(entry->userid, p+19+ITEMID_LEN, (size_t)(tp - (p+19+ITEMID_LEN)));
    entry->userid[tp - (p+19+ITEMID_LEN)] = 0;
    tp= strchr(indextypelist,p[20+ITEMID_LEN+USERID_MAXLEN]);
    if (!tp) goto x_badentry;
    entry->type= tp-indextypelist;
    if (p[21+ITEMID_LEN+USERID_MAXLEN] != ' ') goto x_badentry;
    tp= q-1;
    while (tp > p+22+ITEMID_LEN+USERID_MAXLEN && tp[-1] == ' ') --tp;
    memcpy(entry->subject, p+22+ITEMID_LEN+USERID_MAXLEN,
           (size_t)(tp - (p+22+ITEMID_LEN+USERID_MAXLEN)));
    entry->subject[tp - (p+22+ITEMID_LEN+USERID_MAXLEN)]= 0;
    if ((entry->type == RG_IET_MoTD) != (entry->itemid[0] == 0))
      goto x_badentry;
    entry->lastentrythisitem= (1<<RG_IET_Max)-1;
    for (centry= (RG_IndexEntry*)connP->indexdata.data, i= connP->indexdata.used-1;
         i > 0;
         centry++, i--)
      if (!strcmp(entry->itemid,centry->itemid))
        centry->lastentrythisitem &= ~(1<<entry->type);
    p= q;
  }
  ckfree(p);
  RG_EArrayEnsureSize(&connP->indexorder,connP->indexdata.used,sizeof(RG_IndexEntry*));
  RG_IndexModify_ResetCmd(interp,connP,0);
  connP->indexsince= since;
  connP->indexvalid= level;
  return TCL_OK;
  
 x_badentry: /* error handler */
  Tcl_AppendResult(interp,
                   "server sent bad index entry:\n", p, (char*)0);
  Tcl_DStringFree(&ds);
  return TCL_ERROR;
}
     

